import unittest
from mistake.runtime.interpreter import Interpreter
from mistake.runtime.environment import Environment
from mistake.runtime.runtime_types import RuntimeNumber, RuntimeString, RuntimeBoolean, RuntimeUnit, Function

from mistake.parser.ast import (
    Number,
    String,
    Boolean,
    FunctionDeclaration,
    FunctionApplication,
    VariableDeclaration,
    VariableAccess,
    Block,
    Unit,
)

def reset_wrapper(func):
    def wrapper(self, *args, **kwargs):
        self.interpreter._reset()
        return func(self, *args, **kwargs)
    return wrapper

class TestInterpreter(unittest.TestCase):
    def setUp(self):
        self.interpreter = Interpreter()

    @reset_wrapper
    def test_visit_number(self):
        node = Number(42)
        result = self.interpreter.visit_node(node, self.interpreter.global_environment)
        self.assertIsInstance(result, RuntimeNumber)
        self.assertEqual(result.value, 42)

    @reset_wrapper
    def test_visit_string(self):
        node = String("hello")
        result = self.interpreter.visit_node(node, self.interpreter.global_environment)
        self.assertIsInstance(result, RuntimeString)
        self.assertEqual(result.value, "hello")

    @reset_wrapper
    def test_visit_boolean(self):
        node = Boolean(True)
        result = self.interpreter.visit_node(node, self.interpreter.global_environment)
        self.assertIsInstance(result, RuntimeBoolean)
        self.assertEqual(result.value, True)

    @reset_wrapper
    def test_visit_unit(self):
        node = Unit()
        result = self.interpreter.visit_node(node, self.interpreter.global_environment)
        self.assertIsInstance(result, RuntimeUnit)

    @reset_wrapper
    def test_visit_variable_declaration(self):
        node = VariableDeclaration("x", Number(10), "inf")
        result = self.interpreter.visit_node(node, self.interpreter.global_environment)
        self.assertIsInstance(result, RuntimeNumber)
        self.assertEqual(result.value, 10)
        self.assertEqual(self.interpreter.global_environment.get_variable("x").value, 10)

    @reset_wrapper
    def test_visit_variable_access(self):
        self.interpreter.global_environment.add_variable("x", RuntimeNumber(10), self.interpreter.get_lifetime("inf"))
        node = VariableAccess("x")
        result = self.interpreter.visit_node(node, self.interpreter.global_environment)
        self.assertIsInstance(result, RuntimeNumber)
        self.assertEqual(result.value, 10)

    @reset_wrapper
    def test_visit_function_declaration(self):
        node = FunctionDeclaration("f", ["x"], Block([Number(42)]), False)
        result = self.interpreter.visit_node(node, self.interpreter.global_environment)
        self.assertIsInstance(result, Function)

    @reset_wrapper
    def test_visit_function_application(self):
        func_decl = FunctionDeclaration(["x"], Block([VariableAccess("x")]), impure=True, raw_body="x")
        func = self.interpreter.visit_function_declaration(func_decl, Environment(None))
        self.interpreter.global_environment.add_variable("f", func, self.interpreter.get_lifetime("inf"))
        node = FunctionApplication(VariableAccess("f"), Number(10))
        result = self.interpreter.visit_node(node, self.interpreter.global_environment)
        self.assertIsInstance(result, RuntimeNumber)
        self.assertEqual(result.value, 10)

    @reset_wrapper
    def test_visit_block(self):
        node = Block([VariableDeclaration("x", Number(10), "inf"), VariableAccess("x")])
        result = self.interpreter.visit_node(node, self.interpreter.global_environment)
        self.assertIsInstance(result, RuntimeNumber)
        self.assertEqual(result.value, 10)

    @reset_wrapper
    def test_execute(self):
        ast = [VariableDeclaration("x", Number(10), "inf"), VariableAccess("x")]
        result = self.interpreter.execute(ast)
        self.assertEqual(len(result), 2)
        self.assertIsInstance(result[1], RuntimeNumber)
        self.assertEqual(result[1].value, 10)

if __name__ == "__main__":
    unittest.main()


class RuntimeError(Exception):
    pass


class VariableNotFoundError(RuntimeError):
    pass


class VariableAlreadyDefinedError(RuntimeError):
    pass


class InvalidAssignmentError(RuntimeError):
    pass


class InvalidFunctionCallError(RuntimeError):
    pass


class InvalidLifetimeError(RuntimeError):
    pass


class LifetimeExpiredError(RuntimeError):
    pass


class StackEmptyError(RuntimeError):
    pass


class FileSwitchException(RuntimeError):
    pass


class RuntimeTypeError(RuntimeError):
    pass



import re
import socket
import time
from datetime import datetime
from enum import Enum
from typing import Any, Callable, List, Optional

import gevent

from mistake.parser.ast import ASTNode
from mistake.runtime import environment as rte
from mistake.runtime import interpreter as rt
from mistake.runtime.errors.runtime_errors import InvalidLifetimeError
from mistake.utils import to_decimal_seconds


class MLType:
    def to_string(self):
        raise NotImplementedError("to_string method not implemented")

    def __str__(self):
        return self.to_string()

    def __repr__(self):
        return self.to_string()


class MLCallable(MLType):
    def __call__(self, arg: MLType, env: "rte.Environment", runtime) -> MLType:
        raise NotImplementedError("Callable must implement __call__ method")


class RuntimeUnit(MLType):
    def __init__(self, _=None):
        pass

    def to_string(self):
        return "Null"


class RuntimeNumber(MLType):
    def __init__(self, value: float):
        self.value: float = value

    def to_string(self):
        return f"{self.value}"


class RuntimeString(MLType):
    def __init__(self, value):
        self._value: str = str(value)

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, new_value):
        self._value = str(new_value)

    def to_string(self):
        return f"{self.value}"


class RuntimeBoolean(MLType):
    def __init__(self, value: bool | str):
        self.value: bool = value in ["true", "True", True]

    def to_string(self):
        return "true" if self.value else "false"

    def __eq__(self, other):
        ov = other.value if isinstance(other, RuntimeBoolean) else other
        return self.value == ov


class Function(MLType):
    def __init__(
        self,
        parameter: str,
        body: ASTNode | List,
        captured_env: "rte.Environment",
        impure: bool = True,
        raw_body="",
        is_unparsed=False,
    ):
        self.param: str = parameter
        self.body: ASTNode | List = body

        self.impure: bool = impure
        self.raw_body: str = raw_body
        self.is_unparsed = is_unparsed
        self.captured_env = captured_env

    def to_string(self):
        return f"{'Impure' if self.impure else ''}Function({self.param}, body={self.body}, is_unparsed={self.is_unparsed})"


class ClassType(MLType):
    def __init__(self, members: dict[str, MLType], public_members: set[str]):
        self.members = members
        self.public_members = public_members

    def to_string(self):
        return f"Class(fields={self.members}, public_fields={self.public_members})"


class ClassInstance(MLType):
    def __init__(
        self,
        class_type: ClassType,
        members: dict[str, MLType],
        environment: "rte.Environment",
    ):
        self.class_type = class_type
        self.members = members

        self.environment = environment
        # print("CLASS INSTANCE ENVIRONMENT", self.environment)

    def to_string(self):
        return f"ClassInstance(fields={self.members})"


class BuiltinFunction(MLType):
    def __init__(
        self,
        func: Callable[[MLType, "rte.Environment", "rt.Interpreter"], MLType],
        imp=True,
        subtype=None,
    ):
        self.func = func
        self.impure = imp
        self.subtype = subtype
        self.subdata = {}

    def __call__(self, *args):
        return self.func(*args)

    def to_string(self):
        return f"BuiltinFunction({self.func}, impure={self.impure})"


class LifetimeType(Enum):
    DMS_TIMESTAMP = 0  # given in miliseconds since jan 1st 2020
    DECIMAL_SECONDS = 1
    LINES = 2
    INFINITE = 3
    TICKS = 4


def get_timestamp():
    reference_date = datetime(2020, 1, 1)
    current_time = datetime.now()
    time_difference = current_time - reference_date
    milliseconds = int(time_difference.total_seconds() * 1000)
    decimal_milliseconds = int(to_decimal_seconds(milliseconds))
    return decimal_milliseconds


class Lifetime(MLType):
    def __init__(self, type: LifetimeType, value: float, start: float = 0):
        self.value = value
        self.type = type
        self.start = start

        if self.value < 0:
            raise InvalidLifetimeError("Lifetime value must be greater than 0")
        if int(self.value) != self.value and self.type in [
            LifetimeType.LINES,
            LifetimeType.DMS_TIMESTAMP,
        ]:
            raise InvalidLifetimeError("Lifetime value must be an integer")
        elif int(self.value) == self.value:
            self.value = int(self.value)

    def to_string(self):
        return f"Lifetime({self.value}, {self.type})"

    def is_expired(self, line=None):
        match self.type:
            case LifetimeType.INFINITE:
                return False
            case LifetimeType.DMS_TIMESTAMP:
                return get_timestamp() >= self.value
            case LifetimeType.DECIMAL_SECONDS:
                return (
                    to_decimal_seconds(time.process_time()) - self.start >= self.value
                )  # 0.864 is the conversion from seconds to decimal seconds
            case LifetimeType.TICKS:
                return time.process_time() * 20 - self.start >= self.value
            case LifetimeType.LINES:
                if line is None:
                    raise InvalidLifetimeError(
                        "Line number must be provided for line lifetime"
                    )
                return line - self.start >= self.value


class RuntimeMutableBox(MLType):
    def __init__(self, value: Any):
        self.value = value

    def to_string(self):
        return f"MutableBox({self.value}, {id(self)})"

    def write(self, value):
        self.value = value
        return RuntimeUnit()


class RuntimeListType(MLType):
    def __init__(self, inp: List[MLType] = []):
        self.list = {i + 1: inp[i] for i in range(len(inp))}

    def get(self, idx: int):
        if idx < 1 or int(idx) != idx:
            raise IndexError("Index must be an integer greater than 0")
        return self.list.get(idx, RuntimeUnit())

    def continuous(self) -> List[MLType]:
        return [self.list.get(i, RuntimeUnit()) for i in range(1, len(self.list) + 1)]

    def length(self):
        i = 0
        while (i + 1) in self.list:
            i += 1
        return RuntimeNumber(i)

    def set(self, idx: int, value: MLType):
        if idx < 1 or int(idx) != idx:
            raise IndexError("Index must be an integer greater than 0")
        self.list[idx] = value
        return RuntimeUnit()

    def to_string(self):
        return f"List([{', '.join([str(i) for i in self.list.values()])}])"


def convert_type(val: Any):
    if isinstance(val, dict):
        return runtime_dictify(val)
    elif isinstance(val, list):
        return RuntimeListType([convert_type(v) for v in val])
    elif isinstance(val, str):
        return RuntimeString(val)
    elif isinstance(val, int) or isinstance(val, float):
        return RuntimeNumber(val)
    elif isinstance(val, bool):
        return RuntimeBoolean(val)
    else:
        return val


def un_convert_type(val: Any):
    if isinstance(val, RuntimeDictType):
        return de_runtime_dictify(val)
    elif isinstance(val, RuntimeListType):
        return [un_convert_type(v) for v in val.continuous()]
    elif isinstance(val, RuntimeString):
        return val.value
    elif isinstance(val, RuntimeNumber):
        return val.value
    elif isinstance(val, RuntimeBoolean):
        return val.value
    else:
        return val


def runtime_dictify(d: dict):
    new_d = {}
    for k, v in d.items():
        new_d[convert_type(k)] = convert_type(v)

    return new_d


class RuntimeDictType(MLType):
    def __init__(self, inp: dict = {}):
        self.dict = {k: inp[k] for k in inp}

    def get(self, key: str):
        return self.dict.get(key, RuntimeUnit())

    def set(self, key: str, value: MLType):
        self.dict[key] = value
        return RuntimeUnit()

    def as_regular_dict(self):
        return de_runtime_dictify(self)

    def to_string(self):
        return f"Dict({self.dict}, {id(self)})"


def de_runtime_dictify(d: RuntimeDictType):
    def convert_type(val: MLType) -> Any:
        if isinstance(val, RuntimeDictType):
            return de_runtime_dictify(val)
        elif isinstance(val, RuntimeListType):
            return [convert_type(v) for v in val.continuous()]
        elif isinstance(val, RuntimeString):
            return val.value
        elif isinstance(val, RuntimeNumber):
            return val.value
        elif isinstance(val, RuntimeBoolean):
            return val.value
        else:
            return val

    new_d = {}

    for k, v in d.dict.items():
        new_d[convert_type(k)] = convert_type(v)

    return new_d


class RuntimeMatchObject(MLType):
    def __init__(self, match: re.Match):
        self.match = match

    def to_string(self):
        return f"MatchObject({self.match})"


class RuntimeChannel(MLType):
    def __init__(
        self,
        id: int,
        sent_callback: callable = lambda _: None,
        recieve_callback: callable = lambda: None,
    ):
        self.id = id
        self.sent_callback: callable = sent_callback
        self.receive_callback: callable = recieve_callback
        self.values: List[MLType] = []

    def send(self, value: MLType):
        self.values.append(value)
        self.sent_callback(value)

    def receive(self):
        self.receive_callback()

        while len(self.values) == 0:
            # print(self.channels[id])
            gevent.sleep(0.01)
        return self.values.pop(0)

    def to_string(self):
        return f"Channel({self.id})"


class RuntimeTask(MLType):
    def __init__(self, task_ref: gevent.Greenlet):
        self.task_ref = task_ref
        self.start_time = time.time()

    def kill(self):
        if self.task_ref:
            self.task_ref.kill()

    def to_string(self):
        return f"Task({self.task_ref})"


# NETWORKING
# ADD NEW CLASSES HERE


class RuntimeServer:
    def set_callback(self, callback: callable):
        raise NotImplementedError("Must implement set_callback method")


class RuntimeSocket:
    buffer_size = 1024

    def send(self, message: str):
        raise NotImplementedError("Must implement send method")

    def close(self):
        raise NotImplementedError("Must implement close method")

    def receive(self):
        raise NotImplementedError("Must implement recieve method")


class RuntimeUDPServer(RuntimeServer):
    def __init__(self, runtime):
        self.runtime = runtime
        self.hostname: str = None
        self.port: int = None
        self.socket: socket.socket = None
        self.listen_task = None
        self.callback: callable = None
        self.callback_task = None

    def set_hostname(self, hostname):
        hostname, port = hostname.value.split(":")

        self.hostname = hostname
        self.port = int(port)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.socket.bind((self.hostname, self.port))

        def listen():
            while True:
                try:
                    self.receive()
                except gevent._socketcommon.cancel_wait_ex:
                    break
                gevent.sleep(0.01)

        self.listen_task = gevent.spawn(listen)

        self.runtime.add_task(self.listen_task)

        return RuntimeBoolean(True)

    def receive(self, buffer_size=1024):
        if not self.socket:
            return RuntimeBoolean(False)
        data = self.socket.recvfrom(buffer_size)[0]
        message = data.decode()
        if self.callback:
            task = gevent.spawn(self.callback, RuntimeString(message))
            self.runtime.add_task(task)
            self.callback_task = task

    def set_callback(self, callback: callable):
        self.callback = callback
        return RuntimeUnit()

    def close(self):
        if self.socket:
            self.socket.close()
            self.socket = None
        if self.listen_task:
            self.listen_task.kill()
            self.listen_task = None
        if self.callback_task:
            self.callback_task.kill()
            self.callback_task = None
        return RuntimeBoolean(True)

    def kill(self):
        self.close()

    def to_string(self):
        return f"UDPServer(hostname={self.hostname}, port={self.port})"


class RuntimeUDPSocket(RuntimeSocket):
    def __init__(self, runtime):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.runtime = runtime
        self.hostname = None
        self.port: int = None

    def set_hostname(self, hostname: str):
        self.hostname, self.port = hostname.value.split(":")

        self.socket.connect((self.hostname, int(self.port)))

    def send(self, message: RuntimeString):
        if not self.socket:
            return RuntimeBoolean(False)
        value = message.value.encode()
        self.socket.send(value)
        return RuntimeBoolean(True)

    def close(self):
        self.socket.close()
        return RuntimeBoolean(True)

    def kill(self):
        self.close()

    def receive(self):
        return RuntimeUnit()

    def to_string(self):
        return f"UDPSocket(id={self.socket.fileno()})"


class RuntimeTCPServer(RuntimeServer):
    def __init__(self, runtime):
        self.runtime = runtime
        self.hostname: str = None
        self.port: int = None
        self.socket: socket.socket = None
        self.listen_task = None
        self.callback: callable = None
        self.callback_task = None

    def bind_port(self, port: RuntimeString):
        self.hostname = "127.0.0.1"
        self.port = int(port.value)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.socket.bind((self.hostname, self.port))
        self.socket.listen(5)

        def listen():
            while True:
                try:
                    self.receive()
                except gevent._socketcommon.cancel_wait_ex:
                    break

        self.listen_task = gevent.spawn(listen)
        self.runtime.add_task(self.listen_task)

        return RuntimeBoolean(True)

    def receive(self, buffer_size=1024):
        if not self.socket:
            return RuntimeBoolean(False)

        client_socket, client_address = self.socket.accept()

        new_sock = RuntimeTCPSocket(self.runtime)
        new_sock.set_socket(client_socket)

        if self.callback:
            self.callback_task = gevent.spawn(self.callback, new_sock)
            self.runtime.add_task(self.callback_task)
        return RuntimeUnit()

    def set_callback(self, callback: callable):
        self.callback = callback
        return RuntimeUnit()

    def close(self):
        if self.socket:
            self.socket.close()
            self.socket = None
        if self.listen_task:
            self.listen_task.kill()
            self.listen_task = None
        if self.callback_task:
            self.callback_task.kill()
            self.callback_task = None
        return RuntimeBoolean(True)

    def kill(self):
        self.close()

    def to_string(self):
        return f"TCPServer(hostname={self.hostname}, port={self.port})"


class RuntimeTCPSocket(RuntimeSocket):
    def __init__(self, runtime):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.channel: Optional[RuntimeChannel] = None
        self.runtime = runtime
        self.hostname: Optional[str] = None
        self.port: Optional[int] = None
        self.closed = False

    def set_port(self, port: RuntimeString):
        if self.closed:
            return RuntimeBoolean(False)
        try:
            self.hostname = socket.gethostname()
            self.port = int(port.value)
            self.socket.connect((self.hostname, self.port))
            return RuntimeBoolean(True)
        except Exception:
            self.close()
            return RuntimeBoolean(False)

    def set_socket(self, socket: socket.socket):
        self.socket = socket
        self.closed = False

    def send(self, message: RuntimeString):
        if self.closed:
            return RuntimeBoolean(False)

        if not self.socket:
            return RuntimeBoolean(False)
        try:
            value = message.value.encode()
            self.socket.send(value)
            return RuntimeBoolean(True)
        except Exception as e:
            print(f"Error sending message: {e}")
            return RuntimeBoolean(False)

    def close(self):
        if self.closed:
            return RuntimeBoolean(False)
        try:
            self.socket.close()
            if self.channel:
                self.runtime.close_channel(self.channel)
            self.closed = True
            return RuntimeBoolean(True)
        except Exception as e:
            print(f"Error closing socket: {e}")
            return RuntimeBoolean(False)

    def kill(self):
        self.close()

    def receive(self):
        if self.closed:
            return RuntimeUnit()
        try:
            data = self.socket.recv(1024)
            return RuntimeString(data.decode())
        except Exception as e:
            print(f"Error receiving data: {e}")
            return RuntimeString("")

    def to_string(self):
        return f"TCPSocket(hostname={self.hostname}, port={self.port})"


# AIRTABLE


class RuntimeAirtableBase(MLCallable):
    def __init__(self, base):
        self.base = base

    def __call__(
        self, table_name: RuntimeString, env: "rte.Environment", runtime
    ) -> MLType:
        from mistake.runtime.stdlib.airtable_api import create_table

        # get table instance
        return create_table(self, table_name.value)

    def to_string(self):
        return f"AirtableBase({self.base.name}, {self.base.id})"


class RuntimeAirtableTable(MLType):
    def __init__(self, table):
        self.table = table

    def to_string(self):
        return f"AirtableTable({self.table.id}, {self.table.base.id})"


class RuntimeAirtableRecord(MLType):
    def __init__(self, record: dict):
        self.id = record["id"]
        self.creation_time = record["createdTime"]
        self.fields: dict = record["fields"]

    def set_id(self, id: str):
        self.id = convert_type(id)
        return RuntimeUnit()

    def set_field(self, field: MLType, value: Any):
        self.fields[un_convert_type(field)] = un_convert_type(value)
        return RuntimeUnit()

    def get_field(self, field: MLType):
        # print("GETTING FIELD", field)
        return convert_type(self.fields.get(field, RuntimeUnit()))

    def to_record_dict(self) -> dict:
        return self.fields

    def to_string(self):
        return f"AirtableRecord({self.id}, {self.creation_time}, {self.fields})"



import mistake.runtime.runtime_types as rt
from pyairtable import Base, Api
import requests
from mistake.runtime.errors.runtime_errors import RuntimeError

AIRTABLE_API: Api = None
API_KEY = None


def create_airtable_api_instance(key: "rt.RuntimeString", *_):
    global AIRTABLE_API, API_KEY
    if not isinstance(key, rt.RuntimeString):
        raise rt.RuntimeTypeError(f"Expected RuntimeString, got {type(key)}")

    AIRTABLE_API = Api(key.value)
    API_KEY = AIRTABLE_API.api_key
    return rt.RuntimeUnit()


def create_base(base_id: "rt.RuntimeString"):
    if isinstance(base_id, rt.RuntimeString):
        base_id = base_id.value
    return rt.RuntimeAirtableBase(Base(API_KEY, base_id))


def create_table(base: "rt.RuntimeAirtableBase", table_id: str):
    if isinstance(table_id, rt.RuntimeString):
        table_id = table_id.value
    return rt.RuntimeAirtableTable(AIRTABLE_API.table(base.base.id, table_id))


def list_table_records(table: "rt.RuntimeAirtableTable"):
    if table.table is None:
        raise RuntimeError("Table not found")
    try:
        a = table.table.all()
        return rt.RuntimeListType([rt.RuntimeAirtableRecord(record) for record in a])
    except Exception as e:
        raise e


def get_record(table: "rt.RuntimeAirtableTable", record_id: str):
    return rt.RuntimeAirtableRecord(table.table.get(record_id))


def create_record(table: "rt.RuntimeAirtableTable", record: "rt.RuntimeAirtableRecord"):
    result = table.table.create(record.to_record_dict())
    record.id = result["id"]
    record.creation_time = result["createdTime"]
    return rt.RuntimeUnit()


def update_record(table: "rt.RuntimeAirtableTable", record: "rt.RuntimeAirtableRecord"):
    table.table.update(record.id, record.to_record_dict())
    return rt.RuntimeUnit()


def delete_record(table: "rt.RuntimeAirtableTable", record_id: str):
    table.table.delete(record_id)
    return rt.RuntimeUnit()


def new_record(fields: dict):
    return rt.RuntimeAirtableRecord(
        record={"id": None, "createdTime": None, "fields": fields}
    )


def create_new_field(
    raw_table: "rt.RuntimeAirtableTable",
    field_name: "rt.RuntimeString",
    field_type: "rt.RuntimeString",
    raw_options: "rt.RuntimeDictType",
):
    table = raw_table.table
    # base = table.base
    field_name = rt.un_convert_type(field_name)
    field_type = rt.un_convert_type(field_type)
    options = raw_options.as_regular_dict()
    resp = table.create_field(name=field_name, field_type=field_type, options=options)
    return rt.RuntimeString(resp.model_dump()["id"])


def update_field(
    raw_table: "rt.RuntimeAirtableTable",
    field_id: "rt.RuntimeString",
    new_data: "rt.RuntimeDictType",
):
    table = raw_table.table
    base = table.base
    field_id = rt.un_convert_type(field_id)

    # Endpoint URL
    url = f"https://api.airtable.com/v0/meta/bases/{base.id}/tables/{table.id}/fields/{rt.un_convert_type(field_id)}"

    # Headers
    headers = {"Authorization": f"Bearer {API_KEY}", "Content-Type": "application/json"}

    # Data payload
    data = new_data.as_regular_dict()

    # PATCH request
    response = requests.patch(url, json=data, headers=headers)
    if response.status_code != 200:
        raise RuntimeError(f"Failed to update field: {response.text}")
    return rt.RuntimeUnit()


def all_bases():
    return rt.RuntimeListType(
        [rt.RuntimeAirtableBase(base) for base in AIRTABLE_API.bases()]
    )


def base_schema(base: "rt.RuntimeAirtableBase"):
    return rt.RuntimeDictType(rt.runtime_dictify(base.base.schema().model_dump()))



from mistake.runtime.runtime_types import (
    RuntimeUDPServer,
    RuntimeUDPSocket,
    RuntimeTCPServer,
    RuntimeTCPSocket,
)


def create_UDP_server(arg, env, runtime):
    return RuntimeUDPServer(runtime)


def create_UDP_socket(arg, env, runtime):
    return RuntimeUDPSocket(runtime)


def create_TCP_server(arg, env, runtime):
    return RuntimeTCPServer(runtime)


def create_TCP_socket(arg, env, runtime):
    return RuntimeTCPSocket(runtime)



import re
import time
import gevent
import mistake.runtime.interpreter as interpreter

from mistake.parser.ast import (
    FunctionApplication,
)
from mistake.runtime.errors.runtime_errors import (
    RuntimeTypeError,
)
from mistake.runtime.runtime_types import (
    RuntimeBoolean,
    RuntimeNumber,
    RuntimeString,
    RuntimeUnit,
    RuntimeMutableBox,
    RuntimeListType,
    RuntimeMatchObject,
    RuntimeTask,
    RuntimeDictType,
    BuiltinFunction,
    Function,
    MLType,
    convert_type,
    de_runtime_dictify,
    runtime_dictify,
)
from mistake.runtime.stdlib.networking import (
    create_TCP_server,
    create_UDP_server,
    create_TCP_socket,
    create_UDP_socket,
)
from mistake.utils import from_decimal_seconds, RUNTIME_USE_VULKAN

if RUNTIME_USE_VULKAN:
    import mistake.runtime.stdlib.vulkan_api as vstd
import os

from mistake.runtime.stdlib.airtable_api import (
    all_bases,
    create_airtable_api_instance,
    create_base,
    list_table_records,
    get_record,
    create_record,
    update_record,
    delete_record,
    new_record,
    base_schema,
    create_new_field,
    update_field,
)

from typing import Any, Callable


def get_type(val: Any):
    if isinstance(val, bool):
        return RuntimeBoolean(val)
    if isinstance(val, int) or isinstance(val, float):
        return RuntimeNumber(val)
    if isinstance(val, str):
        return RuntimeString(val)
    if callable(val):
        return BuiltinFunction(val)


_STACK = []
PRINT_CALLBACK = None

def set_print_callback(callback: Callable[[str], None]):
    global PRINT_CALLBACK
    PRINT_CALLBACK = callback
    

def try_pop(arg, env, runtime: "interpreter.Interpreter"):
    if len(_STACK) == 0:
        return RuntimeUnit()

    val = _STACK.pop()

    runtime.visit_function_application(
        env, FunctionApplication(val, arg), visit_arg=False
    )

    return RuntimeUnit()


def get_cur_line(rt: "interpreter.Interpreter"):
    return RuntimeNumber(rt.current_line)


def write_to_mut_box(box: RuntimeMutableBox, *_):
    if not isinstance(box, RuntimeMutableBox):
        raise RuntimeTypeError(f"Expected mutable box, got {type(box)}")
    return BuiltinFunction(lambda arg, *_: box.write(arg), True)


def get_length(arg: MLType, *_):
    match arg:
        case RuntimeString(value):
            # Length of string in bytes
            return RuntimeNumber(len(value.encode()))
        case RuntimeListType(value):
            return RuntimeNumber(len(value))
        case RuntimeNumber(value):
            return len(str(value))
        case RuntimeMatchObject(value):
            return RuntimeNumber(len(value.groups()))
        case RuntimeTask(_, start_time):
            return RuntimeNumber(time.time() - start_time)
        case _:
            return RuntimeNumber(0)


def create_regex_func(arg: RuntimeString, *_):
    try:
        comp = re.compile(arg.value)
        return BuiltinFunction(
            lambda arg, *_: RuntimeListType(
                {
                    (i + 1): RuntimeMatchObject(m)
                    for i, m in enumerate(comp.findall(arg.value))
                }
            ),
            imp=False,
        )
    except re.error:
        return RuntimeUnit()


def get_group_from_match(arg: RuntimeMatchObject, *_):
    return BuiltinFunction(lambda x, *_: RuntimeString(arg.group(x.value)), imp=False)


def get_string_from_match(arg: RuntimeMatchObject, *_):
    return RuntimeString(str(arg.match))


def new_task_from_function_app(
    function: Function, env, runtime: "interpreter.Interpreter", delay: float = 0.0
):
    def task():
        gevent.sleep(from_decimal_seconds(delay))
        runtime.visit_function_application(
            env, FunctionApplication(function, RuntimeUnit()), visit_arg=False
        )

    spawn = gevent.spawn(task)
    runtime.add_task(spawn)
    return RuntimeTask(spawn)


def new_task_from_func(
    func: callable, runtime: "interpreter.Interpreter", delay: float = 0.0
):
    def task():
        gevent.sleep(from_decimal_seconds(delay))
        func()

    spawn = gevent.spawn(task)
    runtime.add_task(spawn)
    return RuntimeTask(spawn)


def is_truthy(arg: MLType, *_):
    if isinstance(arg, RuntimeBoolean):
        return arg.value
    return not isinstance(arg, RuntimeUnit)


def assert_true(arg: MLType, *_):
    if not is_truthy(arg):
        raise AssertionError("Assertion failed")
    return RuntimeUnit()

def mlprint(arg: MLType, *_):
    if PRINT_CALLBACK:
        PRINT_CALLBACK(arg.to_string())
    else:
        print(arg)
    return RuntimeUnit()
    

std_funcs = {
    "?!": BuiltinFunction(lambda arg, *_: mlprint(arg)),
    "+": BuiltinFunction(
        lambda arg, *_: BuiltinFunction(
            lambda x, *_: get_type(arg.value + x.value), imp=False
        ),
        imp=False,
    ),
    "*": BuiltinFunction(
        lambda arg, *_: BuiltinFunction(
            lambda x, *_: get_type(arg.value * x.value), imp=False
        ),
        imp=False,
    ),
    "-": BuiltinFunction(
        lambda arg, *_: BuiltinFunction(
            lambda x, *_: get_type(arg.value - x.value), imp=False
        ),
        imp=False,
    ),
    "/": BuiltinFunction(
        lambda arg, *_: BuiltinFunction(
            lambda x, *_: get_type(arg.value / x.value), imp=False
        ),
        imp=False,
    ),
    "%": BuiltinFunction(
        lambda arg, *_: BuiltinFunction(
            lambda x, *_: get_type(arg.value % x.value), imp=False
        ),
        imp=False,
    ),
    "=": BuiltinFunction(
        lambda arg, *_: BuiltinFunction(
            lambda x, *_: get_type(arg.to_string() == x.to_string())
        ),
        imp=False,
    ),
    ">": BuiltinFunction(
        lambda arg, *_: BuiltinFunction(
            lambda x, *_: get_type(arg.value > x.value), imp=False
        ),
        imp=False,
    ),
    "≥": BuiltinFunction(
        lambda arg, *_: BuiltinFunction(
            lambda x, *_: get_type(arg.value >= x.value), imp=False
        ),
        imp=False,
    ),
    "<": BuiltinFunction(
        lambda arg, *_: BuiltinFunction(
            lambda x, *_: get_type(arg.value < x.value), imp=False
        ),
        imp=False,
    ),
    "≤": BuiltinFunction(
        lambda arg, *_: BuiltinFunction(
            lambda x, *_: get_type(arg.value <= x.value), imp=False
        ),
        imp=False,
    ),
    "≠": BuiltinFunction(
        lambda arg, *_: BuiltinFunction(
            lambda x, *_: get_type(arg.to_string() != x.to_string()), imp=False
        ),
        imp=False,
    ),
    "->": BuiltinFunction(lambda arg, *_: get_length(arg), imp=False),
    "[?]": BuiltinFunction(lambda arg, env, runtime: get_cur_line(runtime), imp=False),
    "|<|": BuiltinFunction(lambda arg, *_: _STACK.append(arg), imp=True),
    "|>|": BuiltinFunction(
        lambda arg, env, runtime: try_pop(arg, env, runtime), imp=True
    ),
    "!": BuiltinFunction(lambda arg, *_: RuntimeMutableBox(arg), imp=True),
    "!<": BuiltinFunction(lambda arg, *_: write_to_mut_box(arg), imp=True),
    "!?": BuiltinFunction(lambda arg, *_: arg.value, imp=True),
    "/?/": BuiltinFunction(create_regex_func, imp=False),
    "/>?/": BuiltinFunction(get_group_from_match, imp=False),
    '/>"/': BuiltinFunction(get_string_from_match, imp=False),
    "??": BuiltinFunction(lambda arg, *_: RuntimeString(arg.to_string()), imp=False),
    "!!": BuiltinFunction(lambda arg, *_: assert_true(arg), imp=False),
    # Lists
    "[!]": BuiltinFunction(lambda *args: RuntimeListType(), imp=False),
    "[<]": BuiltinFunction(
        lambda arg, *_: BuiltinFunction(
            lambda index, *_: BuiltinFunction(
                lambda value, *_: arg.set(index.value, value), imp=False
            ),
            imp=False,
        )
    ),
    "[>]": BuiltinFunction(
        lambda arg, *_: BuiltinFunction(lambda x1, *_: arg.get(x1.value), imp=False)
    ),
    "[/]": BuiltinFunction(
        lambda arg0, *_: BuiltinFunction(
            lambda arg1, env, runtime: new_task_from_function_app(
                arg1, env, runtime, arg0.value
            ),
            imp=True,
        )
    ),
    "<!>": BuiltinFunction(
        lambda arg, env, runtime: new_task_from_function_app(arg, env, runtime, 0),
        imp=False,
    ),
    "</>": BuiltinFunction(lambda arg, env, runtime: arg.kill(), imp=False),
    "=!=": BuiltinFunction(
        lambda arg, env, runtime: runtime.create_channel(), imp=True
    ),  # Create a channel
    "<<": BuiltinFunction(
        lambda arg, env, runtime: BuiltinFunction(
            lambda x, *_: runtime.send_to_channel(arg, x), imp=True
        )
    ),  # Send to channel
    ">>": BuiltinFunction(
        lambda arg, env, runtime: runtime.receive_from_channel(arg), imp=True
    ),  # Receive from channel
    # NETWORKING
    #'<=#=>': BuiltinFunction(lambda arg, env, runtime: create_TCP_server(arg, env, runtime), imp=True),
    "<=?=>": BuiltinFunction(
        lambda arg, env, runtime: create_UDP_server(arg, env, runtime), imp=True
    ),
    "<=#=>": BuiltinFunction(
        lambda arg, env, runtime: create_TCP_server(arg, env, runtime), imp=True
    ),
    "<=?=": BuiltinFunction(
        lambda arg, env, runtime: create_UDP_socket(arg, env, runtime), imp=True
    ),
    "<=#=": BuiltinFunction(
        lambda arg, env, runtime: create_TCP_socket(arg, env, runtime), imp=True
    ),
    "==>#": BuiltinFunction(
        lambda x0, *_: BuiltinFunction(
            lambda x1, env, runtime: x0.bind_port(x1), imp=True
        )
    ),
    "==>?": BuiltinFunction(
        lambda x0, *_: BuiltinFunction(lambda x1, env, runtime: x0.set_hostname(x1)),
        imp=True,
    ),
    "==>!": BuiltinFunction(
        lambda x0, *_: BuiltinFunction(
            lambda x1, env, runtime: x0.set_callback(
                lambda callback_arg: runtime.visit_function_application(
                    env, FunctionApplication(x1, callback_arg), visit_arg=False
                )
            ),
            imp=True,
        )
    ),
    ">|<": BuiltinFunction(lambda arg, *_: arg.close(), imp=True),
    # DICTIONARIES
    "{+}": BuiltinFunction(lambda *_: RuntimeDictType(), imp=False),
    ">{}": BuiltinFunction(
        lambda d, *_: BuiltinFunction(
            lambda key, *_: BuiltinFunction(lambda value, *_: d.set(key, value))
        ),
        imp=False,
    ),
    "<{}": BuiltinFunction(
        lambda d, *_: BuiltinFunction(lambda key, *_: d.get(key), imp=False)
    ),
    # MISC HELPERS
    '>"<': BuiltinFunction(lambda arg, *_: RuntimeString(arg.value.strip()), imp=False),
    # AIRTABLE
    "{>!<}": BuiltinFunction(
        lambda arg, *_: create_airtable_api_instance(arg)
    ),  # create airtable api
    "{!}": BuiltinFunction(lambda arg, *_: create_base(arg)),  # create base
    "{?}": BuiltinFunction(
        lambda table, *_: list_table_records(table)
    ),  # get all records
    "{>}": BuiltinFunction(
        lambda table, *_: BuiltinFunction(lambda id, *_: get_record(table, id))
    ),  # get record
    "{<}": BuiltinFunction(
        lambda table, *_: BuiltinFunction(
            lambda record, *_: create_record(table, record)
        )
    ),  # put record
    "{!": BuiltinFunction(
        lambda fields, *_: new_record(de_runtime_dictify(fields))
    ),  # create local record instance
    "{-}": BuiltinFunction(
        lambda table, *_: BuiltinFunction(
            lambda record_id, *_: delete_record(table, record_id.value)
        )
    ),  # delete record
    "{\\}": BuiltinFunction(
        lambda table, *_: BuiltinFunction(
            lambda record, *_: update_record(table, record)
        )
    ),  # update record
    # Schema editing
    "{{?": BuiltinFunction(
        lambda table, *_: RuntimeDictType(runtime_dictify(table.table.schema().dict()))
    ),  # get schema
    "{{+": BuiltinFunction(
        lambda table, *_: BuiltinFunction(
            lambda field, *_: BuiltinFunction(
                lambda field_type, *_: BuiltinFunction(
                    lambda options, *_: create_new_field(
                        table, field, field_type, options
                    )
                )
            )
        )
    ),  # add field    # Record editing
    "{{=": BuiltinFunction(
        lambda table, *_: BuiltinFunction(
            lambda field_id, *_: BuiltinFunction(
                lambda new_options, *_: update_field(table, field_id, new_options)
            )
        )
    ),  # update field
    # RECORD EDITING
    "{<": BuiltinFunction(
        lambda record, *_: BuiltinFunction(
            lambda field, *_: BuiltinFunction(
                lambda value, *_: record.set_field(field, value)
            )
        )
    ),  # create airtable api
    "{>": BuiltinFunction(
        lambda record, *_: BuiltinFunction(lambda field, *_: field.fields[record])
    ),  # create airtable api
    "{#>": BuiltinFunction(lambda record, *_: convert_type(record.id)),  # get ID
    "{#<": BuiltinFunction(
        lambda record, *_: BuiltinFunction(lambda ID, *_: record.set_id(ID))
    ),  # set ID
    # BASE MANIPULATION
    "{}?": BuiltinFunction(lambda *args: all_bases(*args)),
    "{}??": BuiltinFunction(lambda base, *_: base_schema(base)),
    # ENV
    "[@@@]": BuiltinFunction(lambda key, *_: RuntimeString(os.getenv(key.value))),
}

if RUNTIME_USE_VULKAN:
    std_funcs.update(vstd.vulkan_std_funcs)



import re
import sys

RUNTIME_USE_VULKAN = "--vulkan" in sys.argv 

def is_latin_alph(c: str) -> bool:
    return re.fullmatch(r"[a-zA-Z]", c) is not None


def to_decimal_seconds(seconds: float):
    return seconds * 0.864


def from_decimal_seconds(seconds: float):
    return seconds / 0.864

ansi_color = {
    "red": "\033[91m",
    "green": "\033[92m",
    "yellow": "\033[93m",
    "blue": "\033[94m",
    "magenta": "\033[95m",
    "cyan": "\033[96m",
    "white": "\033[97m",
    "black": "\033[30m",
    "reset": "\033[0m"
}

def print_color(color: str, *args):
    print(f"{ansi_color[color]}{' '.join(map(str, args))}{ansi_color['reset']}")


import time
from typing import List

# Must keep this down here to supress warnings
from gevent import monkey
monkey.patch_all()

from mistake import runner # noqa: E402
from mistake.parser.ast import ( # noqa: E402
    ASTNode,
    Block,
    Boolean,
    ClassDefinition,
    ClassInstancing,
    FunctionApplication,
    FunctionDeclaration,
    JumpStatement,
    Match,
    MemberAccess,
    Number,
    String,
    Unit,
    VariableAccess,
    VariableDeclaration,
) 
from mistake.parser.parser import Parser # noqa: E402
from mistake.runtime.environment import ContextType, Environment # noqa: E402
from mistake.runtime.errors.runtime_errors import RuntimeError # noqa: E402
from mistake.runtime.runtime_types import (  # noqa: E402
    BuiltinFunction,
    ClassInstance,
    ClassType,
    Function,
    Lifetime,
    LifetimeType,
    MLCallable,
    MLType,
    RuntimeBoolean,
    RuntimeChannel,
    RuntimeNumber,
    RuntimeString,
    RuntimeUnit,
    get_timestamp,
) 

from mistake.utils import to_decimal_seconds # noqa: E402

import gevent  # noqa: E402


class Interpreter:
    def __init__(self, unsafe_mode=False):
        self.unsafe_mode = unsafe_mode
        self.parser = Parser()
        self.global_environment = Environment(None, context_type=ContextType.IMPURE)
        self.current_line = 1
        self.files: dict[str, List[ASTNode]] = {}
        self.tasks: List[gevent.Greenlet] = []
        self.channel_id = 0
        self.current_line = 1
        self.lines_executed = 1

    def _reset(self):
        self.global_environment = Environment(None, context_type=ContextType.IMPURE)
        self.current_line = 1
        self.files: dict[str, List[ASTNode]] = {}
        self.tasks: List[gevent.Greenlet] = []
        self.channel_id = 0
        self.current_line = 1
        self.lines_executed = 1

    def create_channel(self, cb_s=lambda *_: None, cb_r=lambda: None):
        self.channel_id += 1
        return RuntimeChannel(self.channel_id, cb_s, cb_r)

    def send_to_channel(self, channel: RuntimeChannel, value: MLType):
        channel.send(value)
        return RuntimeUnit()

    def receive_from_channel(self, channel: RuntimeChannel):
        return channel.receive()

    def run_all_tasks(self):
        if self.tasks:
            # Run tasks asynchronously without blocking the main thread
            for task in self.tasks:
                task.start()
            self.tasks = [task for task in self.tasks if not task.ready()]

    def add_task(self, task: gevent.Greenlet):
        self.tasks.append(task)

    def visit_function_application(
        self,
        env: Environment,
        node: FunctionApplication,
        visit_arg: bool = True,
    ):
        param = self.visit_node(node.parameter, env) if visit_arg else node.parameter

        function = self.visit_node(node.called, env)
        is_builtin = False

        if not isinstance(function, Function):
            if not isinstance(function, BuiltinFunction) and not isinstance(
                function, MLCallable
            ):
                raise RuntimeError(
                    f"Called {node.called} is not a function, but a {type(function)}"
                )
            else:
                is_builtin = True

        if function.impure and env.context_type == ContextType.PURE:
            raise RuntimeError(
                f"Function {function} is impure and cannot be called in a pure context"
            )

        if is_builtin:
            result = function(param, env, self)
            return result

        new_env = Environment(
            env,
            context_type=ContextType.IMPURE if function.impure else ContextType.PURE,
        )

        if function.captured_env is not None:
            new_env.absorb_environment(function.captured_env)
            new_env = Environment(
                new_env,
                context_type=ContextType.IMPURE
                if function.impure
                else ContextType.PURE,
            )

        new_env.add_variable(function.param, param, Lifetime(LifetimeType.INFINITE, 0))

        return self.visit_node(function.body, new_env)

    def visit_function_declaration(self, node: FunctionDeclaration, env: Environment):
        params = node.parameters

        cap_env = Environment(None, context_type=env.context_type)
        for var in env.get_all_defined_vars():
            value, life = env.get_full_var_data(var)
            cap_env.add_variable(
                var,
                value,
                life,
                ignore_duplicate=True,
            )

        # curry the function
        def get_curried(params, body):
            if len(params) == 1:
                return Function(
                    params[0],
                    body,
                    is_unparsed=True,
                    raw_body=node.raw_body,
                    impure=node.impure,
                    captured_env=cap_env,
                )
            return Function(
                params[0],
                Block([get_curried(params[1:], body)]),
                is_unparsed=False,
                raw_body=node.raw_body,
                impure=node.impure,
                captured_env=cap_env,
            )

        r = get_curried(params, node.body)
        return r

    def visit_block(self, node: Block, env: Environment):
        new_env = Environment(env, context_type=env.context_type)
        if not node.body:
            return Unit()

        for statement in node.body[:-1]:
            self.visit_node(statement, new_env)

        return self.visit_node(node.body[-1], new_env)

    def get_lifetime(self, lifetime: str, *_):
        if lifetime == "inf":
            return Lifetime(LifetimeType.INFINITE, 0)
        
        lifetime_value = int(lifetime[:-1])
        lifetime_type = lifetime[-1]
        
        if lifetime_type == 'l':
            return Lifetime(LifetimeType.LINES, lifetime_value, self.lines_executed)
        elif lifetime_type == 'u':
            return Lifetime(LifetimeType.DMS_TIMESTAMP, lifetime_value, get_timestamp())
        elif lifetime_type == 't':
            return Lifetime(LifetimeType.TICKS, lifetime_value, time.process_time() * 20)
        elif lifetime_type == 's':
            return Lifetime(LifetimeType.DECIMAL_SECONDS, lifetime_value, to_decimal_seconds(time.process_time()))
        else:
            raise RuntimeError(f"Invalid lifetime {lifetime}")

    def visit_class_definition(self, node: ClassDefinition, env: Environment):
        parent_class = None
        members = {}
        pmembers = set()
        if node.parent:
            parent_class = env.get_variable(node.parent, line=self.lines_executed)
            if not isinstance(parent_class, ClassType):
                raise RuntimeError(f"'{node.parent}' is not a valid class.")

            members = {name: value for name, value in parent_class.members.items()}
            pmembers = parent_class.public_members

        for name, value in node.members.items():
            members[name] = value

        pmembers.update(node.public_members)

        new_class = ClassType(members, pmembers)
        return new_class

    def visit_class_instancing(self, node: ClassInstancing, env: Environment):
        # Lookup the class in the environment
        class_type = env.get_variable(node.name, line=self.lines_executed)
        if not isinstance(class_type, ClassType):
            raise RuntimeError(f"'{node.name}' is not a valid class.")

        # Create a new instance with the class fields
        instance_members = {name: value for name, value in class_type.members.items()}

        instance_env = Environment(env, context_type=ContextType.IMPURE)
        for name, value in instance_members.items():
            v = self.visit_node(value, instance_env)
            if isinstance(v, Function):
                v.captured_env = instance_env
            instance_env.add_variable(
                name,
                v,
                Lifetime(LifetimeType.INFINITE, 0),
            )  # HACK: No lifetime handling for instance members because that's stupid

        return ClassInstance(class_type, instance_members, instance_env)

    def visit_member_access(self, node: MemberAccess, env: Environment):
        # Lookup the instance in the environment
        instance = self.visit_node(node.obj, env)

        if node.member == '"':
            if isinstance(instance, Function):
                return RuntimeString(instance.raw_body)

        if not isinstance(instance, ClassInstance):
            raise RuntimeError(f"'{node.obj}' is not a valid instance.")

        # Access the field of the instance
        if node.member not in instance.members:
            raise RuntimeError(f"'{node.member}' is not a valid field of '{node.obj}'.")
        if node.member not in instance.class_type.public_members:
            raise RuntimeError(
                f"'{node.member}' is not a public field of '{node.obj}'."
            )

        return instance.environment.get_variable(node.member, line=self.lines_executed)

    def visit_match(self, node: Match, env: Environment):
        expr = self.visit_node(node.expr, env)
        env.add_variable("@", expr, Lifetime(LifetimeType.INFINITE, 0))
        for case in node.cases:
            v = self.visit_node(case.condition, env)

            if v:
                return self.visit_node(case.expr, env)
        return self.visit_node(node.otherwise, env)

    def visit_node(self, node: ASTNode, env: Environment, imperative=False):
        if isinstance(node, Unit):
            return RuntimeUnit()
        if isinstance(node, Number):
            return RuntimeNumber(node.value)
        if isinstance(node, String):
            return RuntimeString(node.value)
        if isinstance(node, Boolean):
            return RuntimeBoolean(node.value)
        if isinstance(node, Function):
            return node
        if isinstance(node, VariableAccess):
            return env.get_variable(node.name, line=self.lines_executed)
        if isinstance(node, FunctionApplication):
            return self.visit_function_application(env, node)
        if isinstance(node, Block):
            return self.visit_block(node, env)
        if isinstance(node, VariableDeclaration):
            value = self.visit_node(node.value, env)
            env.add_variable(node.name, value, self.get_lifetime(node.lifetime, node))
            return value
        if isinstance(node, FunctionDeclaration):
            return self.visit_function_declaration(node, env)
        if isinstance(node, ClassDefinition):
            return self.visit_class_definition(node, env)
        if isinstance(node, ClassInstancing):
            return self.visit_class_instancing(node, env)
        if isinstance(node, MemberAccess):
            return self.visit_member_access(node, env)
        if isinstance(node, Match):
            return self.visit_match(node, env)
        if isinstance(node, JumpStatement):
            self.swap_file(
                self.visit_node(node.file_expr, env),
                self.visit_node(node.line_expr, env),
            )
            return Unit()
        raise NotImplementedError(f"Node {node} not implemented")

    def swap_file(self, filename: MLType, line: int):
        if isinstance(filename, RuntimeString):
            filename = filename.value
        else:
            raise RuntimeError(f"Expected string, got {filename}")

        if isinstance(line, RuntimeNumber):
            line = int(line.value)
        else:
            raise RuntimeError(f"Expected number, got {line}")

        if filename not in self.files:
            self.files[filename] = runner.fetch_file(filename)

        self.ast = self.files[filename]
        self.current_line = line - 1
        if self.current_line > len(self.ast):
            raise RuntimeError(f"Line {line} is out of bounds in file {filename}")

    def execute(
        self, ast: List[ASTNode], filename: str = "NO_FILE", standalone=False
    ) -> List[MLType]:
        self.ast = ast
        self.current_line = 1
        self.lines_executed = 1
        self.files[filename] = ast

        result: List[MLType] = []

        while self.current_line <= len(self.ast):
            node = self.ast[self.current_line - 1]
            try:
                result.append(
                    self.visit_node(node, self.global_environment, imperative=True)
                )
                self.run_all_tasks()

            except Exception as e:
                if self.unsafe_mode and not standalone:
                    raise e
                if not standalone:
                    print(f"Error at line {self.current_line}, {e}")
                return result + [e]
            self.current_line += 1
            self.lines_executed += 1

        gevent.joinall(self.tasks)
        return result



from typing import List

from mistake.parser.ast import (
    ASTNode,
    Boolean,
    JumpStatement,
    Unit,
    VariableAccess,
    String,
    Number,
    Block,
    FunctionDeclaration,
    FunctionApplication,
    VariableDeclaration,
    Match,
    MatchCase,
    ClassDefinition,
    MemberAccess,
    ClassInstancing,
)
from mistake.parser.errors.parser_errors import ParserError, UnexpectedTokenError
from mistake.tokenizer.lexer import Lexer
from mistake.tokenizer.token import Token, TokenType, opening_tokens


class Parser:
    breaking_tokens = [
        TokenType.KW_END,
        TokenType.KW_CLOSE,
        TokenType.SYM_EOF,
        TokenType.KW_CASES,
        TokenType.KW_THEN,
        TokenType.KW_OF,
        TokenType.KW_DO,
        TokenType.KW_FROM,
    ]

    def __init__(self):
        self.tokens = []
        self.position = 0
        self.current_token = None
        self.errors: List[ParserError] = []

    def get_current(self):
        if self.position < len(self.tokens):
            return self.tokens[self.position]
        raise ParserError("Unexpected end of input")
    
    def parse(self, tokens: List[Token]) -> ASTNode:
        self.tokens = tokens

        self.position = 0
        self.current_token = self.get_current()

        return self.parse_program()

    def synchronize(self):
        while self.current_token.type not in self.breaking_tokens:
            self.advance()
    def parse_program(self):
        nodes = []
        while self.current_token.type != TokenType.SYM_EOF:
            try:
                nodes.append(self.parse_node())
            except (ParserError, UnexpectedTokenError) as e:
                self.errors.append(e)
                try:
                    self.advance()
                    self.synchronize()
                except ParserError:
                    break
        return nodes

    def eat(self, token_type: TokenType):
        start = self.position
        if self.current_token.type == token_type:
            self.advance()
            return self.tokens[start]
        else:
            raise UnexpectedTokenError(
                f"Unexpected token {self.current_token} at line {self.current_token.line}, expected {token_type}"
            )

    def advance(self, allow_ws=False):
        self.position += 1
        while (
            self.get_current().type == TokenType.SYM_WHITESPACE and not allow_ws
        ):
            self.position += 1
        if self.position < len(self.tokens):
            self.current_token = self.get_current()

    def peek_next_is(self, token_type: TokenType):
        return self.tokens[self.position + 1].type == token_type

    def next_is(self, token_type: TokenType):
        return self.current_token.type == token_type

    def parse_node(self):
        if self.current_token.type in [TokenType.KW_PUBLIC, TokenType.KW_VARIABLE]:
            val = self.parse_variable_declaration()
        elif self.current_token.type == TokenType.KW_JUMP:
            val = self.parse_jump_statement()
        else:
            try:
                val = self.parse_expression()
            except UnexpectedTokenError as e:
                raise e
        self.eat(TokenType.KW_END)
        return val

    def parse_single_expr(self, atom: ASTNode, allow_function_application=True):
        if (
            self.current_token.type in self.breaking_tokens
            or not allow_function_application
        ):
            return atom

        return self.parse_function_application(atom)

    def parse_expression(self, allow_function_application=True):
        match self.current_token.type:
            case TokenType.SYM_NUMBER:
                val = Number(self.eat(TokenType.SYM_NUMBER).value)
            case TokenType.SYM_STRING:
                val = String(self.eat(TokenType.SYM_STRING).value)
                self.eat(TokenType.KW_CLOSE)
            case TokenType.KW_TRUE:
                self.eat(TokenType.KW_TRUE)
                val = Boolean(True)
            case TokenType.KW_FALSE:
                self.eat(TokenType.KW_FALSE)
                val = Boolean(False)
            case TokenType.KW_UNIT:
                self.eat(TokenType.KW_UNIT)
                val = Unit()
            case TokenType.SYM_IDENTIFIER:
                val = self.parse_id_expression(
                    allow_function_application=allow_function_application
                )
            case TokenType.KW_OPEN:
                val = self.parse_block()
            case TokenType.KW_FUNCTION:
                val = self.parse_function_declaration()
            case TokenType.KW_IMPURE:
                val = self.parse_function_declaration()
            case TokenType.KW_MATCH:
                val = self.parse_match_expression()
            case TokenType.KW_CLASS:
                val = self.parse_class_definition()
            case TokenType.KW_MEMBER:
                val = self.parse_member_access()
            case TokenType.KW_NEW:
                val = self.parse_class_instancing()
            case TokenType.KW_USE:
                val = self.parse_use()
            case TokenType.KW_WITH:
                val = self.parse_with()
            case _:
                raise UnexpectedTokenError(
                    f"Unexpected token {self.current_token} at line {self.current_token.line}"
                )

        return self.parse_single_expr(
            val, allow_function_application=allow_function_application
        )


    def parse_variable_declaration(self):
        public = False
        if self.next_is(TokenType.KW_PUBLIC):
            self.eat(TokenType.KW_PUBLIC)
            public = True
        self.eat(TokenType.KW_VARIABLE)
        name = self.eat(TokenType.SYM_IDENTIFIER).value
        if self.next_is(TokenType.KW_TYPE):
            self.eat(TokenType.KW_TYPE)
            self.advance()
        lifetime = "inf"
        if self.next_is(TokenType.KW_LIFETIME):
            self.eat(TokenType.KW_LIFETIME)
            lifetime = self.eat(TokenType.SYM_DURATION).value

        self.eat(TokenType.KW_IS)
        value = self.parse_expression()
        return VariableDeclaration(name, value, lifetime, public)

    def parse_case(self):
        self.eat(TokenType.KW_CASE)
        condition = self.parse_expression()
        self.eat(TokenType.KW_THEN)
        body = self.parse_expression()
        self.eat(TokenType.KW_CLOSE)
        return MatchCase(condition, body)

    def parse_match_expression(self):
        # match <expression> cases [case <expression> then <expression> close]mistake..? otherwise <expression> close

        self.eat(TokenType.KW_MATCH)
        expr = self.parse_expression()
        self.eat(TokenType.KW_CASES)
        cases = []
        while self.current_token.type != TokenType.KW_OTHERWISE:
            cases.append(self.parse_case())

        self.eat(TokenType.KW_OTHERWISE)
        otherwise = self.parse_expression()
        self.eat(TokenType.KW_CLOSE)
        self.eat(TokenType.KW_CLOSE)
        return Match(expr, cases, otherwise)

    def get_unparsed_body(self):
        body: List[Token] = []
        stack = 1
        while stack > 0:
            if self.current_token.type in opening_tokens:
                stack += 1
            if self.current_token.type == TokenType.KW_CLOSE:
                stack -= 1
            if stack == 0:
                break

            body.append(self.current_token)
            self.advance(allow_ws=True)

        if body[0].type != TokenType.KW_OPEN:
            body = (
                [Token(TokenType.KW_OPEN, "open")]
                + body
                + [Token(TokenType.KW_CLOSE, "close")]
            )

        return body + [Token(TokenType.KW_END, "end"), Token(TokenType.SYM_EOF, "eof")]

    def parse_class_definition(self):
        self.eat(TokenType.KW_CLASS)
        inherit = None
        if self.next_is(TokenType.KW_INHERITS):
            self.eat(TokenType.KW_INHERITS)
            inherit = self.eat(TokenType.SYM_IDENTIFIER).value
        self.eat(TokenType.KW_HAS)

        members = {}
        pmembers = set()
        body: List[VariableDeclaration] = []
        while self.current_token.type != TokenType.KW_CLOSE:
            body.append(v := self.parse_variable_declaration())
            if v.public:
                pmembers.add(v.name)
            self.eat(TokenType.KW_END)

        for member in body:
            members[member.name] = member.value

        self.eat(TokenType.KW_CLOSE)
        return ClassDefinition(members, inherit, pmembers)

    def parse_member_access(self):
        self.eat(TokenType.KW_MEMBER)
        member = self.eat(TokenType.SYM_IDENTIFIER).value
        self.eat(TokenType.KW_OF)
        obj = self.parse_expression()
        return MemberAccess(obj, member)

    def parse_function_declaration(self):
        impure = False
        if self.next_is(TokenType.KW_IMPURE):
            self.eat(TokenType.KW_IMPURE)
            impure = True

        self.eat(TokenType.KW_FUNCTION)
        parameters = []
        while self.current_token.type != TokenType.KW_RETURNS:
            parameters.append(self.eat(TokenType.SYM_IDENTIFIER).value)

        self.eat(TokenType.KW_RETURNS)
        body = self.get_unparsed_body()
        np = Parser()
        parsed_body = np.parse(body)[0]
        self.eat(TokenType.KW_CLOSE)
        
        def get_curried_function(params: List[str], body: ASTNode):
            if not params:
                return body
            return FunctionDeclaration([params[0]], get_curried_function(params[1:], body), impure=impure)
        
        return get_curried_function(parameters, parsed_body)

    def parse_use(self):
        self.eat(TokenType.KW_USE)
        parameters: List[str] = []
        while self.current_token.type != TokenType.KW_FROM:
            parameters.append(self.eat(TokenType.SYM_IDENTIFIER).value)
        self.eat(TokenType.KW_FROM)
        call_expr = self.parse_expression()
        self.eat(TokenType.KW_DO)
        body_expr = self.parse_expression()
        
        callback = FunctionDeclaration(parameters, [Block([body_expr])], impure=True, is_unparsed=False)
        return FunctionApplication(call_expr, callback)

    def parse_with(self):
        self.eat(TokenType.KW_WITH)
        raw = [self.parse_expression()]
        while self.current_token.type != TokenType.KW_CLOSE:
            self.eat(TokenType.KW_DO)
            raw.append(self.parse_expression())
        self.eat(TokenType.KW_CLOSE)
        
        def build_nested_call(exprs: List[ASTNode]):
            if len(exprs) == 1:
                return exprs[0]
            return FunctionApplication(exprs.pop(-1), build_nested_call(exprs))
            
        res = build_nested_call(raw.copy())
        return res
        
    def parse_id_expression(self, allow_function_application=True):
        if (
            self.tokens[self.position + 1].type not in self.breaking_tokens
            and allow_function_application
        ):
            return self.parse_function_application(
                VariableAccess(self.eat(TokenType.SYM_IDENTIFIER).value)
            )

        return VariableAccess(self.eat(TokenType.SYM_IDENTIFIER).value)

    def get_function_application_from_params(
        self, expr: ASTNode, params: List[ASTNode]
    ):
        for param in params:
            expr = FunctionApplication(expr, param)
        return expr

    def parse_function_application(self, func: ASTNode = None):
        parameters = []

        while self.current_token.type not in self.breaking_tokens:
            parameters.append(self.parse_expression(allow_function_application=False))

        return self.get_function_application_from_params(func, parameters)

    def is_final_expr_in_block(self):
        stack = [
            self.tokens[self.position].type
        ]  # Initialize with the current opening token

        for i in range(self.position + 1, len(self.tokens)):
            token = self.tokens[i]

            if token.type in opening_tokens:
                stack.append(token.type)
            elif token.type == TokenType.KW_CLOSE:
                stack.pop()

                if not stack:
                    # We've closed all nested blocks
                    return True
            elif token.type == TokenType.KW_END and len(stack) == 1:
                # If we find an 'end' at the current block depth, it's not the final expression
                return False

        # If we exit the loop without returning, we've hit the end of the input
        return True

    def parse_block(self):
        self.eat(TokenType.KW_OPEN)
        body = []
        while self.current_token.type != TokenType.KW_CLOSE:
            # print("PARSING BLOCK", self.current_token, self.is_final_expr_in_block())
            if self.is_final_expr_in_block():
                body.append(self.parse_expression())
            else:
                body.append(self.parse_node())
        self.eat(TokenType.KW_CLOSE)
        return Block(body)

    def parse_class_instancing(self):
        self.eat(TokenType.KW_NEW)
        name = self.eat(TokenType.SYM_IDENTIFIER).value
        return ClassInstancing(name)

    def parse_jump_statement(self):
        self.eat(TokenType.KW_JUMP)
        line = self.parse_expression()
        self.eat(TokenType.KW_OF)
        file = self.parse_expression()
        return JumpStatement(file, line)


def get_file_ast(file: str) -> List[ASTNode]:
    with open(file) as f:
        code = f.read()
        lexer = Lexer()
        parser = Parser()
        tokens = lexer.tokenize(code)
        ast = parser.parse(tokens)
        return ast



import time

from mistake.runtime.errors.runtime_errors import (
    LifetimeExpiredError,
    VariableAlreadyDefinedError,
    VariableNotFoundError,
)
from mistake.runtime.runtime_types import Lifetime, MLType
from mistake.runtime.stdlib import std_funcs as stdlib


class ContextType:
    PURE = 0
    IMPURE = 1


test = []


class Environment:
    def __init__(
        self, parent: "Environment", context_type: ContextType = ContextType.IMPURE
    ):
        self.variables: dict[str, MLType] = {}
        self.lifetimes: dict[str, Lifetime] = {}
        self.parent = parent
        self.context_type = context_type
        self.test_time = time.process_time_ns()
        test.append(self)

    def get_variable(self, name: str, line: int = 0) -> MLType:
        if name in self.variables:
            if self.lifetimes[name].is_expired(line):
                del self.variables[name]
                del self.lifetimes[name]
                raise LifetimeExpiredError(f"Lifetime for variable {name} has expired")
            return self.variables[name]

        if self.parent:  # and self.context_type == ContextType.IMPURE:
            return self.parent.get_variable(name)

        if name in stdlib.std_funcs:
            return stdlib.std_funcs[name]

        raise VariableNotFoundError(f"Variable {name} not found.")

    def get_full_var_data(self, name: str):
        if name in self.variables:
            return self.variables[name], self.lifetimes[name]
        if self.parent:
            return self.parent.get_full_var_data(name)
        raise VariableNotFoundError(f"Variable {name} not found.")

    def add_variable(
        self, name: str, value: MLType, lifetime: Lifetime, ignore_duplicate=False
    ):
        if name == "_":
            return

        if not ignore_duplicate and name in self.variables:
            raise VariableAlreadyDefinedError(
                f"Variable {name} already defined in this scope"
            )

        self.variables[name] = value

        if not isinstance(lifetime, Lifetime):
            raise TypeError(f"{lifetime} must be of type Lifetime")
        self.lifetimes[name] = lifetime

    def get_all_defined_vars(self):
        vars = []
        for var in self.variables:
            vars.append(var)
        if self.parent:
            vars.extend(self.parent.get_all_defined_vars())
        return vars

    def absorb_environment(self, env: "Environment"):
        for var in env.variables:
            self.add_variable(
                var, env.get_variable(var), env.lifetimes[var], ignore_duplicate=True
            )

    def repr_simple(self):
        return f"{str(id(self))[-4:]}({','.join(self.get_all_defined_vars())})({self.test_time})"

    def __repr__(self):
        out = f"Environment({id(self)}, (\n"
        for var in self.variables:
            out += f"   {var}: {self.variables[var].to_string()}\n"

        out += f"), {self.context_type}, {id(self.parent)})"
        return out



from enum import Enum
from typing import Any


class TokenType(Enum):
    ERROR = -1
    KW_VARIABLE = 0
    KW_IS = 1
    KW_OPEN = 2
    KW_CLOSE = 3
    KW_IMPURE = 9

    SYM_IDENTIFIER = 11
    KW_STRING = 12

    SYM_OPEN_PAREN = 14
    SYM_CLOSE_PAREN = 15
    KW_END = 16
    KW_RETURNS = 17

    SYM_WHITESPACE = 19
    SYM_DURATION = 20
    KW_LIFETIME = 21

    KW_PUBLIC = 24
    KW_RANDOMIZE = 25
    KW_TYPE = 26

    # TYPES
    TYPE_NUMBER = 27
    TYPE_BOOLEAN = 29

    # LITERALS
    KW_UNIT = 31
    KW_TRUE = 22
    KW_FALSE = 23
    KW_FUNCTION = 10
    SYM_STRING = 18
    SYM_NUMBER = 13

    KW_JUMP = 30
    KW_NEW = 32
    KW_CLASS = 33
    KW_INHERITS = 34
    KW_HAS = 35
    KW_MATCH = 36
    KW_CASES = 38
    KW_CASE = 37
    KW_OTHERWISE = 39
    KW_MEMBER = 40
    KW_THEN = 41
    KW_OF = 42
    KW_DO = 43
    KW_USE = 44
    KW_FROM = 45
    KW_WITH = 46
    
    EX_COMMENT = 47
    EX_STRING = 48
    
    SYM_EOF = 100


opening_tokens = [
    TokenType.KW_OPEN,
    TokenType.KW_CASES,
    TokenType.KW_CASE,
    TokenType.KW_OTHERWISE,
    TokenType.KW_RETURNS,
    TokenType.KW_HAS,
    TokenType.SYM_STRING,
]


class Token:
    def __init__(self, type: TokenType, value: Any, line: int = 0):
        self.value = value
        self.type = type
        self.line = line

    def __str__(self) -> str:
        v = self.value
        v = v.replace("\n", "\\n").replace("\r", "\\r").replace("\t", "\\t")
        return f'Token({self.type}: "{v}")'

    def __repr__(self) -> str:
        return str(self)



from typing import List


class ASTNode:
    def __init__(self):
        pass

    def __str__(self):
        pass

    def __repr__(self):
        return self.__str__()

    def to_string(self):
        return self.__str__()


class VariableAccess(ASTNode):
    def __init__(self, name: str):
        self.name = name

    def __str__(self):
        return f"VariableAccess({self.name})"


class FunctionApplication(ASTNode):
    def __init__(self, called: ASTNode, parameter: ASTNode):
        self.called = called
        self.parameter = parameter

    def __str__(self):
        return f"FunctionApplication({self.called}, {self.parameter})"


class Number(ASTNode):
    def __init__(self, value: float):
        self.value = float(value)

    def __str__(self):
        return str(self.value)


class String(ASTNode):
    def __init__(self, value: str):
        self.value = value

    def __str__(self):
        return self.value


class Boolean(ASTNode):
    def __init__(self, value: bool):
        self.value = value

    def __str__(self):
        return str(self.value)


class Unit(ASTNode):
    def __init__(self):
        pass

    def __str__(self):
        return "Unit"


class Block(ASTNode):
    def __init__(self, body: List[ASTNode]):
        self.body = body

    def __str__(self):
        return f"Block({self.body})"


class VariableDeclaration(ASTNode):
    def __init__(
        self, name: str, value: ASTNode, lifetime: str = "inf", public: bool = False
    ):
        self.name = name
        self.value = value
        self.lifetime = lifetime
        self.public = public

    def __str__(self):
        return f"VariableDeclaration({self.name}, value={self.value}, lifetime={self.lifetime}, public={self.public})"


class FunctionDeclaration(ASTNode):
    def __init__(
        self, parameters: List[str], body: ASTNode, impure: bool = False, raw_body=""
    ):
        self.parameters = parameters
        self.body = body
        self.impure = impure
        self.raw_body = raw_body
    def __str__(self):
        return f"FunctionDeclaration({self.parameters}, body={self.body}, impure={self.impure})"


class MatchCase(ASTNode):
    def __init__(self, condition: ASTNode, expr: ASTNode):
        self.condition = condition
        self.expr = expr

    def __str__(self):
        return f"MatchCase({self.condition}, {self.expr})"


class Match(ASTNode):
    def __init__(self, expr: ASTNode, cases: List[MatchCase], otherwise: ASTNode):
        self.expr = expr
        self.cases = cases
        self.otherwise = otherwise

    def __str__(self):
        return f"Match({self.expr}, cases={self.cases}, otherwise={self.otherwise})"


class ClassDefinition(ASTNode):
    def __init__(
        self,
        members: dict[str, ASTNode],
        parent: str | None = None,
        pmembers: set[str] = set(),
    ):
        self.members = members
        self.parent = parent
        self.public_members = pmembers

    def __str__(self):
        return f"ClassDefinition({self.members}, inherit={self.parent})"


class MemberAccess(ASTNode):
    def __init__(self, obj: str, member: str):
        self.obj = obj
        self.member = member

    def __str__(self):
        return f"MeberAccess({self.obj}, {self.member})"


class ClassInstancing(ASTNode):
    def __init__(self, name: str):
        self.name = name

    def __str__(self):
        return f"ClassInstance({self.name})"


class JumpStatement(ASTNode):
    def __init__(self, file_expr: ASTNode, line_expr: ASTNode):
        self.file_expr = file_expr
        self.line_expr = line_expr

    def __str__(self):
        return f"JumpStatement({self.file_expr}, {self.line_expr})"


import time
import sys
import os
from mistake.tokenizer.lexer import Lexer
from mistake.parser.parser import Parser
from mistake.runtime.interpreter import Interpreter
from mistake.runtime.environment import test
from typing import Tuple, List
from dotenv import load_dotenv
import argparse
import mistake.localize as localize
import json

ENV_PATH = None

def print_help():
    print("Usage: mistake-lang <filename> [--time] [--tokens] [--ast] [--no-exe] [--env <path>] [--vulkan]")

def get_args() -> Tuple[str, List[str]]:
    parser = argparse.ArgumentParser(description='MistakeLang Interpreter')
    parser.add_argument('filename', type=str, help='The script file to run')
    parser.add_argument('--time', action='store_true', help='Print timing information')
    parser.add_argument('--tokens', action='store_true', help='Print tokens')
    parser.add_argument('--ast', action='store_true', help='Print AST')
    parser.add_argument('--no-exe', action='store_true', help='Do not execute the script')
    parser.add_argument('--env', type=str, help='Path to the .env file')
    parser.add_argument('--vulkan', action='store_true', help='Enable Vulkan support')
    parser.add_argument('--unsafe', action='store_true', help='Enable unsafe mode')
    parser.add_argument('--end-env', action='store_true', help='Print the global environment at the end')
    parser.add_argument('--language', type=str, help='Language for localization, use lang = "purge" to purge all localizations')
    parser.add_argument('--env-tree', action='store_true', help='Print the environment tree')
    
    args = parser.parse_args()

    if not os.path.isfile(args.filename):
        print(f"File '{args.filename}' not found")
        sys.exit(1)

    return args

def run_script(program: str, lex=None, parser=None, rt=None, standalone=True) -> List:
    if lex is None: 
        lex = Lexer()
    if parser is None: 
        parser = Parser()
    if rt is None: 
        rt = Interpreter()
    
    tokens = lex.tokenize(program)
    ast = parser.parse(tokens)  
    if parser.errors:
        for error in parser.errors:
            print(error)
        return []
    return rt.execute(ast, standalone=standalone)

def get_system_language() -> str:
    return os.getenv('LANG', 'en').split('.')[0]

def graph_env_relations():
    graph = {} # {parent: [children]}
    for env in test:
        graph[str(id(env))[-4:]] = []
    
    for env in test:
        if env.parent:
            graph[str(id(env.parent))[-4:]].append(env.repr_simple())
        
    return graph
def main():
    global ENV_PATH
    args = get_args()
    p_time = args.time

    cur_path = os.path.dirname(os.path.abspath(__file__))


    lexer = Lexer()
    parser = Parser()
    runtime = Interpreter(args.unsafe)

    lang = args.language if args.language is not None else get_system_language()
    if lang == "purge":
        localize.purge_localizations()
        return
    file = os.path.join(cur_path, f"./tokenizer/.localizations/{lang}.json")
    if not os.path.isfile(file):
        print(f"Localization file for {lang} not found, generating new translation.")
        localize.translate(lang)
        
    with open(file, "r", encoding="utf-8") as f:
        translate = json.loads(f.read())
        trans, keywords = translate.keys(), translate.values()
        n = {}
        for t, kw in zip(trans, keywords):
            n[t] = lexer.keywords[kw]
                
        lexer.keywords = n
            

    ENV_PATH = args.env
    if ENV_PATH is not None: 
        load_dotenv(ENV_PATH)
    else: 
        load_dotenv()
    
    with open(args.filename, "r", encoding='utf-8') as f:
        if p_time:
            print("Read file:", time.process_time())
        start = time.time()

        os.chdir(os.path.dirname(os.path.abspath(args.filename)))
        code = f.read()

        tokens = lexer.tokenize(code)
        if p_time:
            print("Tokenized:", time.process_time())

        if args.tokens:
            print(tokens)

        ast = parser.parse(tokens)
        if p_time:
            print("Parsed:", time.process_time())

        if parser.errors:
            for error in parser.errors:
                print(error)
            return

        if args.ast:
            print("AST BEGIN")
            for i in ast:
                print(i)
            print("AST END")

        if not args.no_exe:
            runtime.execute(ast, filename=args.filename)
        
        if args.end_env:
            print(runtime.global_environment)
        
        if p_time:
            print(f"Total runtime: {time.time() - start} seconds")
        
        if args.env_tree:
            print(json.dumps(graph_env_relations(), indent=3)) 
if __name__ == "__main__":  
    main()



import requests
import os
import json
from typing import List
from mistake.tokenizer.lexer import keywords_en
from concurrent.futures import ThreadPoolExecutor

GOOGLE_TRANSLATE_URL = "https://translate.googleapis.com/translate_a/single"

def translate_keyword(keyword: str, dest_language: str) -> str:
    params = {
        "client": "gtx",
        "sl": "en",  # Source language (English)
        "tl": dest_language,  # Target language
        "dt": "t",
        "q": keyword
    }

    response = requests.get(GOOGLE_TRANSLATE_URL, params=params)
    if response.status_code != 200:
        print(f"Failed to translate {keyword}. HTTP {response.status_code}")
        return keyword
    translated_text = response.json()[0][0][0]
    return translated_text


def translate_keywords(keywords: List[str], dest_language: str) -> List[str]:
    def translate_single_keyword(keyword):
        print(f"Translating {keyword}...")
        return translate_keyword(keyword, dest_language)

    with ThreadPoolExecutor() as executor:
        result = list(executor.map(translate_single_keyword, keywords))
        
    return result

def purge_localizations():
    localizations_path = os.path.join(os.path.dirname(__file__), "./tokenizer/.localizations")
    if os.path.exists(localizations_path):
        for file in os.listdir(localizations_path):
            file_path = os.path.join(localizations_path, file)
            if os.path.isfile(file_path):
                os.remove(file_path)
        print(f"Purged all localization files in {localizations_path}")
        os.rmdir(localizations_path)
    else:
        print(f"No localization files found in {localizations_path}")

def translate(dest_language: str):
    localizations_path = os.path.join(os.path.dirname(__file__), f"./tokenizer/.localizations/{dest_language}.json")
    os.makedirs(os.path.dirname(localizations_path), exist_ok=True)

    english = list(keywords_en.keys())

    translated_keywords = translate_keywords(english, dest_language)

    ret = {}
    for i, j in zip(english, translated_keywords):
        ret[j.lower().replace(" ", "_")] = i.lower()

    with open(localizations_path, "w", encoding="utf-8") as f:
        json.dump(ret, f, ensure_ascii=False, indent=4)

    print(f"Translated keywords saved to {localizations_path}")
    return ret

if __name__ == "__main__":
    translate("es")



from mistake.parser.parser import Parser
from mistake.tokenizer.lexer import Lexer
from typing import List

def fetch_file(filename: str) -> List:
    return Parser().parse(Lexer().tokenize(open(filename).read()))



from mistake.tokenizer.token import Token, TokenType
from mistake.utils import is_latin_alph
from typing import List
from mistake.parser.errors.parser_errors import ParserError
import re
import html

keywords_en: dict[str, TokenType] = {
    "variable": TokenType.KW_VARIABLE,
    "is": TokenType.KW_IS,
    "open": TokenType.KW_OPEN,
    "close": TokenType.KW_CLOSE,
    "impure": TokenType.KW_IMPURE,
    "function": TokenType.KW_FUNCTION,
    "end": TokenType.KW_END,
    "returns": TokenType.KW_RETURNS,
    "lifetime": TokenType.KW_LIFETIME,
    "true": TokenType.KW_TRUE,
    "false": TokenType.KW_FALSE,
    "public": TokenType.KW_PUBLIC,
    "randomize": TokenType.KW_RANDOMIZE,
    "type": TokenType.KW_TYPE,
    "number": TokenType.TYPE_NUMBER,
    "boolean": TokenType.TYPE_BOOLEAN,
    "unit": TokenType.KW_UNIT,
    "jump": TokenType.KW_JUMP,
    "new": TokenType.KW_NEW,
    "class": TokenType.KW_CLASS,
    "inherits": TokenType.KW_INHERITS,
    "has": TokenType.KW_HAS,
    "match": TokenType.KW_MATCH,
    "case": TokenType.KW_CASE,
    "cases": TokenType.KW_CASES,
    "otherwise": TokenType.KW_OTHERWISE,
    "member": TokenType.KW_MEMBER,
    "then": TokenType.KW_THEN,
    "of": TokenType.KW_OF,
    "do": TokenType.KW_DO,
    "use": TokenType.KW_USE,
    "from": TokenType.KW_FROM,
    "with": TokenType.KW_WITH,
    "comment": TokenType.EX_COMMENT,
    "string": TokenType.EX_STRING,    
}
class Lexer:
    def __init__(self, keywords = keywords_en):
        self.errors: List[ParserError] = []
        self.tokens: List[Token] = []
        self.code = ""
        self.current_token = None
        self.current_position = 0
        self.current_line = 1
        self.keywords = keywords

    def add_token(self, token: Token):
        if token.type == TokenType.SYM_WHITESPACE and token.value == "":
            return
        self.tokens.append(token)

    def is_identifier(self, s: str) -> bool:
        contains_non_number = False
        for c in s:
            if not c.isdigit():
                contains_non_number = True
            if is_latin_alph(c):
                return False
        return contains_non_number

    def get_token(self, s: str) -> TokenType:
        if s in self.keywords:
            return self.keywords[s]
        if re.fullmatch(r"\-?[0-9]+(\.[0-9]+)?", s) is not None:
            return TokenType.SYM_NUMBER
        if s[:-1].isdigit() and s[-1] in "lust":
            return TokenType.SYM_DURATION
        if self.is_identifier(s):
            return TokenType.SYM_IDENTIFIER

        return TokenType.ERROR

    def skip_whitespace(self):
        ws = ""
        while (
            self.current_position < len(self.code)
            and self.code[self.current_position].isspace()
        ):
            ws += self.code[self.current_position]
            if self.code[self.current_position] == "\n":
                self.current_line += 1
            self.current_position += 1
        if ws != "":
            self.add_token(Token(TokenType.SYM_WHITESPACE, ws, self.current_line))

    def skip_line(self):
        while (
            self.current_position < len(self.code)
            and self.code[self.current_position] != "\n"
        ):
            self.current_position += 1
        self.current_line += 1
        self.current_position += 1

    def get_string(self) -> str:
        start = self.current_position
        while (
            self.current_position < len(self.code)
            and self.code[self.current_position : self.current_position + 5] != "close"
        ):
            self.current_position += 1

        return html.unescape(self.code[start + 1 : self.current_position - 1]).replace(
            "<br>", "\n"
        )

    def tokenize(self, code: str) -> List[Token]:
        self.code = code
        while self.current_position < len(self.code):
            self.skip_whitespace()
            if self.current_position >= len(self.code):
                break

            start = self.current_position
            while (
                self.current_position < len(self.code)
                and not self.code[self.current_position].isspace()
            ):
                self.current_position += 1

            t = self.code[start : self.current_position]

            token_type = self.get_token(t)
            if token_type == TokenType.EX_COMMENT:
                self.skip_line()
                continue

            if token_type == TokenType.EX_STRING:
                self.add_token(
                    Token(TokenType.SYM_STRING, self.get_string(), self.current_line)
                )
                continue

            if token_type == TokenType.ERROR:
                self.add_token(Token(TokenType.ERROR, t, self.current_line))
            else:
                self.add_token(Token(token_type, t, self.current_line))

        self.add_token(Token(TokenType.SYM_EOF, "EOF", self.current_line))
        
        # remove all prefix whitespace tokens
        while self.tokens[0].type == TokenType.SYM_WHITESPACE:
            self.tokens.pop(0)
        return self.tokens

    def __str__(self):
        return "\n".join([str(token) for token in self.tokens])

    def __repr__(self):
        return self.__str__()



import subprocess
from typing import List, Tuple
from mistake.runtime.runtime_types import MLType, RuntimeListType, BuiltinFunction, RuntimeUnit, Function
import mistake.runtime.stdlib.std_funcs as std
from copy import deepcopy

try:
    import kp
    import numpy as np
except ImportError as e:
    raise ImportError(
        "Mistake GPU support requires additional dependencies. Install them with:\n\n"
        "    pip install mistake-lang[full]\n"
    ) from e


class RuntimeVulkanDevice(MLType):
    def __init__(self):
        pass

    def manager(self):
        return kp.Manager()

    def to_string(self):
        return "VulkanDevice()"

class RuntimeVulkanBuffer(MLType):
    def __init__(self, data: list, dtype: np.dtype):
        self.data = data
        self.dtype = dtype

    def to_string(self):
        return f"VulkanBuffer({self.data}, {repr(self.dtype)})"


def new_vulkan_buffer(arg: RuntimeListType, *_):
    return BuiltinFunction(
        lambda x, *_: RuntimeVulkanBuffer(deepcopy(x).continuous(), arg), imp=False
    )


def actually_run_vulkan_shader(
    threads: Tuple[int, int],
    shader_code: str,
    in_buffers: RuntimeListType,
    out_buffers: RuntimeListType,
    device: RuntimeVulkanDevice,
):
    in_buffs: List[RuntimeVulkanBuffer] = in_buffers.continuous()
    out_buffs: List[RuntimeVulkanBuffer] = out_buffers.continuous()

    def on_finish(result: List[List]):
        for new, cur in zip(result, out_buffs):
            cur.data = [i for i in new]
            
    kompute(threads, shader_code, in_buffs, out_buffs, device, on_finish)

    return RuntimeUnit()    


def run_vulkan_shader(shader_func: Function, *_):
    code = shader_func.body[1:-4]
    code = "".join(i.value for i in code)

    return BuiltinFunction(
        lambda device, *_: BuiltinFunction(
            lambda x_threads, *_: BuiltinFunction(
                lambda y_threads, *_: BuiltinFunction(
                    lambda in_buffers, *_: BuiltinFunction(
                        lambda out_buffers, *_: actually_run_vulkan_shader(
                            (x_threads.value, y_threads.value), code, in_buffers, out_buffers, device
                        ),
                        imp=True,
                    ),
                    imp=False,
                ),
                imp=False,
            ),
            imp=False,
        ),
        imp=False,
    )


def compile_source(shader_code):
    TEMP_DIR = "/tmp"
    # Write the shader code to a temporary file
    with open(f"{TEMP_DIR}/temp_shader.comp", "w") as f:
        f.write(shader_code)

    # Compile the shader using glslangValidator
    result = subprocess.run(
        [
            "glslangValidator",
            "-V",
            f"{TEMP_DIR}/temp_shader.comp",
            "-o",
            f"{TEMP_DIR}/temp_shader.spv",
        ],
        capture_output=True,
        text=True,
    )
    if result.returncode != 0:
        raise RuntimeError(f"Shader compilation failed: {result.stderr}")
    # Read the compiled SPIR-V binary
    with open(f"{TEMP_DIR}/temp_shader.spv", "rb") as f:
        spirv = f.read()
    return spirv


def kompute(
    threads: Tuple[int, int],
    shader,
    raw_in: List[RuntimeVulkanBuffer],
    raw_out: List[RuntimeVulkanBuffer],
    manager: RuntimeVulkanDevice,
    finished_callback=None,
):
    
    # 1. Create Kompute Manager with default settings (device 0, first queue and no extensions)
    mgr = manager.manager()


    # 2. Create and initialise Kompute Tensors through manager
    #print(shader)
    in_tensors = [mgr.tensor_t(np.array([i.value for i in buf.data], dtype=buf.dtype)) for buf in raw_in]
    out_tensors = [
        mgr.tensor_t(np.array([i.value for i in buf.data], dtype=buf.dtype)) for buf in raw_out
    ]

    params = [*in_tensors, *out_tensors]

    # 3. Create algorithm based on shader (supports buffers & push/spec constants)
    workgroup = (int(threads[0]), int(threads[1]), 1)

    # See documentation shader section for compile_source
    spirv = compile_source(shader)

    algo = mgr.algorithm(params, spirv, workgroup)

    # 4. Run operation synchronously using sequence
    (
        mgr.sequence()
        .record(kp.OpTensorSyncDevice(params))
        .record(kp.OpAlgoDispatch(algo))  # Binds default push consts provided
        .eval()
    )  # evaluates only the last recorded op

    # 5. Sync results from the GPU asynchronously
    sq = mgr.sequence()
    sq.eval_async(kp.OpTensorSyncLocal(params))
    sq.eval_await()

    out = [i.data().tolist() for i in out_tensors]

    if finished_callback:
        finished_callback(out)

    return out


if __name__ == "__main__":
    # Define a raw string shader (or use the Kompute tools to compile to SPIRV / C++ header
    # files). This shader shows some of the main components including constants, buffers, etc
    shader = """
        #version 460
    
        layout(local_size_x = 1) in;
        
        layout(set = 0, binding = 0) buffer Data {
            uint data[];
        } buf;
        
        void main() {
            uint idx = gl_GlobalInvocationID.x;
            buf.data[idx] *= 26;
        }
    """

    o = kompute(
        shader,
        [RuntimeVulkanBuffer([1, 2, 3, 4, 5], np.uint32)],
        [],
        RuntimeVulkanDevice(),
    )  # [([125,152,182], np.uint32), ([4,5,6], np.uint32)], [([0,0,0], np.uint32), ([0,0,0], np.uint32)])
    print(o)


vulkan_std_funcs = {
    # VULKAN
    "🔥+32": np.uint32,
    "🔥[!]": BuiltinFunction(lambda *args: new_vulkan_buffer(*args)),
    "🔥🔥": RuntimeVulkanDevice(),
    '🔥🔥()': BuiltinFunction(lambda *args: run_vulkan_shader(*args)),
    '🔥[<]': BuiltinFunction(lambda buffer, *_: RuntimeListType([(std.get_type(i) if not isinstance(i, MLType) else i) for i in buffer.data])),
}